// Image Matrix Converter version 0.1
//
// # License (MIT License)
// Copyright 2018 RIKEN.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// # Author
// This program was written by Yusuke HIROTA.
//
// # Pre-requirements
// - Scilab 5.5.2 or compatible version.
// - Scilab module SIVP.

function [u] = nearest_uint8(r)
    // Return a nearest integer in [0, 255].
    u = uint8(min(max(round(r), 0), 255));
    return u;
endfunction

function quit_button()
    disp("quit_button");
    quit;
endfunction

function select_image_button()
    global message_text;
    disp("select_image_button");
    new_path = uigetfile(["*.*","All *.* files"]);
    disp(new_path);
    if (new_path == "") then
      disp("cancelled.");
      return; // User selected "Cancel". Do nothing.
    else
      target_image = new_path;
      disp("Updated. New target_file = " + target_image);
      message_text.String = "新しい画像を選択しました．"
    end
    im = im2uint8(rgb2gray(imread(target_image)));
    sizexy = size(im);
    scale = 640 / sizexy(1);
    if (scale > 1.0) then
        scale = 1.0;
    end
    im = imresize(im, scale)
    imshow(im);
    [im, message_text, target_image] = resume(im, message_text, target_image); // Write back
endfunction

function set_default_button()
    global message_text;
    message_text.String = "画像を最初の状態に戻しました．";
    disp("set_default_button");
    im = im2uint8(rgb2gray(imread(target_image)));
    sizexy = size(im);
    scale = 640 / sizexy(1);
    if (scale > 1.0) then
        scale = 1.0;
    end
    im = imresize(im, scale)
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function imcomplement_button()
    global message_text;
    message_text.String = "<html>画像の白黒を反転させました．<br/>各画素の新しい濃淡値として，255 と現在の濃淡値の差が設定されています．</html>";
    disp("imcomplement_button");
    im = imcomplement(im);
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function svd_slider_percent_button(slider)
    global message_text;
    disp("svd_slider_percenet_button");
    disp(svd_slider.Value);
    sizexy = size(im);
    tic();
    [U,S,V] = svd(double(im));
    m = int(max(1, round(min(sizexy) * svd_slider.Value * 0.01)));
    disp(m);
    im = nearest_uint8(U(:, 1 : m) * S(1 : m, 1 : m) * (V(:, 1 : m))');
    t = toc();
    disp(t);
    imshow(im);
    message_text.String = "<html>特異値分解を用いた画像の圧縮を行いました．<br/>画像行列 <i>X</i> の特異値分解を行い，<br/>そのうちの一部のみを記憶することでデータサイズを削減しています．<br/>今回の処理では，<i>k</i> = " + string(m) + " 本の特異ベクトルを使用しました．</html>";
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function update_svd_slider(slider)
    global svd_slider_text;
    disp("update slider");
    disp(svd_slider.Value);
    svd_slider_text.String = string(svd_slider.Value * 2) + svd_slider_text_suffix;
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function gaussian_2_button()
    global message_text;
    message_text.String = "<html>ガウスぼかしフィルタを適用しました．<br/>変換行列 <i>A</i> と画像行列 <i><b>x</b></i> の積 <i>A<b>x</b></i> が計算されています．</html>";
    disp("gaussian_2_button");
    sizexy = size(im);
    tic();
    im_16 = uint16(im);

    im2_16 = im_16 * 4;
    im2_16(1 : sizexy(1) - 1, :) = im2_16(1 : sizexy(1) - 1, :) + im_16(2 : sizexy(1), :) * 2;
    im2_16(2 : sizexy(1), :) = im2_16(2 : sizexy(1), :) + im_16(1 : sizexy(1) - 1, :) * 2;
    im2_16(:, 1 : sizexy(2) - 1) = im2_16(:, 1 : sizexy(2) - 1) + im_16(:, 2 : sizexy(2)) * 2;
    im2_16(:, 2 : sizexy(2)) = im2_16(:, 2 : sizexy(2)) + im_16(:, 1 : sizexy(2) - 1) * 2;
    im2_16(1 : sizexy(1) - 1, 1 : sizexy(2) - 1) = im2_16(1 : sizexy(1) - 1, 1 : sizexy(2) - 1) + im_16(2 : sizexy(1), 2 : sizexy(2));
    im2_16(2 : sizexy(1), 1 : sizexy(2) - 1) = im2_16(2 : sizexy(1), 1 : sizexy(2) - 1) + im_16(1 : sizexy(1) - 1, 2 : sizexy(2));
    im2_16(1 : sizexy(1) - 1, 2 : sizexy(2)) = im2_16(1 : sizexy(1) - 1, 2 : sizexy(2)) + im_16(2 : sizexy(1), 1 : sizexy(2) - 1);
    im2_16(2 : sizexy(1), 2 : sizexy(2)) = im2_16(2 : sizexy(1), 2 : sizexy(2)) + im_16(1 : sizexy(1) - 1, 1 : sizexy(2) - 1);

    im2_16(2 : sizexy(1) - 1, 2 : sizexy(2) - 1) = im2_16(2 : sizexy(1) - 1, 2 : sizexy(2) - 1) / 16;
    im2_16(1, 2 : sizexy(2) - 1) = im2_16(1, 2 : sizexy(2) - 1) / 12;
    im2_16(sizexy(1), 2 : sizexy(2) - 1) = im2_16(sizexy(1), 2 : sizexy(2) - 1) / 12;
    im2_16(2 : sizexy(1) - 1, 1) = im2_16(2 : sizexy(1) - 1, 1) / 12;
    im2_16(2 : sizexy(1) - 1, sizexy(2)) = im2_16(2 : sizexy(1) - 1, sizexy(2)) / 12;
    im2_16(1, 1) = im2_16(1, 1) / 9;
    im2_16(1, sizexy(2)) = im2_16(1, sizexy(2)) / 9;
    im2_16(sizexy(1), 1) = im2_16(sizexy(1), 1) / 9;
    im2_16(sizexy(1), sizexy(2)) = im2_16(sizexy(1), sizexy(2)) / 9;
    t = toc();
    disp(t);
    im = nearest_uint8(im2_16)
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function sharpen_9_button()
    // Require sizexy(1) >= 2 and sizexy(2) >= 2.
    global message_text;
    message_text.String = "sharpen_9_button message";
    disp("sharpen_9_button");
    sizexy = size(im);
    tic();
    im_16 = uint16(im);

    im2_16 = im_16 * 9;
    im2_16(1 : sizexy(1) - 1, :) = im2_16(1 : sizexy(1) - 1, :) - im_16(2 : sizexy(1), :);
    im2_16(2 : sizexy(1), :) = im2_16(2 : sizexy(1), :) - im_16(1 : sizexy(1) - 1, :);
    im2_16(:, 1 : sizexy(2) - 1) = im2_16(:, 1 : sizexy(2) - 1) - im_16(:, 2 : sizexy(2));
    im2_16(:, 2 : sizexy(2)) = im2_16(:, 2 : sizexy(2)) - im_16(:, 1 : sizexy(2) - 1);
    im2_16(1 : sizexy(1) - 1, 1 : sizexy(2) - 1) = im2_16(1 : sizexy(1) - 1, 1 : sizexy(2) - 1) - im_16(2 : sizexy(1), 2 : sizexy(2));
    im2_16(2 : sizexy(1), 1 : sizexy(2) - 1) = im2_16(2 : sizexy(1), 1 : sizexy(2) - 1) - im_16(1 : sizexy(1) - 1, 2 : sizexy(2));
    im2_16(1 : sizexy(1) - 1, 2 : sizexy(2)) = im2_16(1 : sizexy(1) - 1, 2 : sizexy(2)) - im_16(2 : sizexy(1), 1 : sizexy(2) - 1);
    im2_16(2 : sizexy(1), 2 : sizexy(2)) = im2_16(2 : sizexy(1), 2 : sizexy(2)) - im_16(1 : sizexy(1) - 1, 1 : sizexy(2) - 1);

    im2_16(2 : sizexy(1) - 1, 2 : sizexy(2) - 1) = im2_16(2 : sizexy(1) - 1, 2 : sizexy(2) - 1) / 1; // Do nothing.
    im2_16(1, 2 : sizexy(2) - 1) = im2_16(1, 2 : sizexy(2) - 1) / 4;
    im2_16(sizexy(1), 2 : sizexy(2) - 1) = im2_16(sizexy(1), 2 : sizexy(2) - 1) / 4;
    im2_16(2 : sizexy(1) - 1, 1) = im2_16(2 : sizexy(1) - 1, 1) / 4;
    im2_16(2 : sizexy(1) - 1, sizexy(2)) = im2_16(2 : sizexy(1) - 1, sizexy(2)) / 4;
    im2_16(1, 1) = im2_16(1, 1) / 6;
    im2_16(1, sizexy(2)) = im2_16(1, sizexy(2)) /  6;
    im2_16(sizexy(1), 1) = im2_16(sizexy(1), 1) /  6;
    im2_16(sizexy(1), sizexy(2)) = im2_16(sizexy(1), sizexy(2)) /  6;
    t = toc();
    disp(t);
    im = nearest_uint8(im2_16)
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function sharpen_5_button()
    // Require sizexy(1) >= 2 and sizexy(2) >= 2.
    global message_text;
    message_text.String = "sharpen_5_button message";
    disp("sharpen_5_button");
    sizexy = size(im);
    tic();
    im_16 = uint16(im);

    im2_16 = im_16 * 5;
    im2_16(1 : sizexy(1) - 1, :) = im2_16(1 : sizexy(1) - 1, :) - im_16(2 : sizexy(1), :);
    im2_16(2 : sizexy(1), :) = im2_16(2 : sizexy(1), :) - im_16(1 : sizexy(1) - 1, :);
    im2_16(:, 1 : sizexy(2) - 1) = im2_16(:, 1 : sizexy(2) - 1) - im_16(:, 2 : sizexy(2));
    im2_16(:, 2 : sizexy(2)) = im2_16(:, 2 : sizexy(2)) - im_16(:, 1 : sizexy(2) - 1);

    im2_16(2 : sizexy(1) - 1, 2 : sizexy(2) - 1) = im2_16(2 : sizexy(1) - 1, 2 : sizexy(2) - 1) / 1; // Do nothing.
    im2_16(1, 2 : sizexy(2) - 1) = im2_16(1, 2 : sizexy(2) - 1) / 2;
    im2_16(sizexy(1), 2 : sizexy(2) - 1) = im2_16(sizexy(1), 2 : sizexy(2) - 1) / 2;
    im2_16(2 : sizexy(1) - 1, 1) = im2_16(2 : sizexy(1) - 1, 1) / 2;
    im2_16(2 : sizexy(1) - 1, sizexy(2)) = im2_16(2 : sizexy(1) - 1, sizexy(2)) / 2;
    im2_16(1, 1) = im2_16(1, 1) / 3;
    im2_16(1, sizexy(2)) = im2_16(1, sizexy(2)) /  3;
    im2_16(sizexy(1), 1) = im2_16(sizexy(1), 1) /  3;
    im2_16(sizexy(1), sizexy(2)) = im2_16(sizexy(1), sizexy(2)) /  3;
    t = toc();
    disp(t);
    im = nearest_uint8(im2_16)
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function dark_button()
    global message_text;
    message_text.String = "<html>画像を暗くしました．<br/>全画素の濃淡値に0.5をかける処理が行われています．</html>";
    disp("dark_button");
    sizexy = size(im);
    tic();
    im2 = zeros(sizexy(1), sizexy(2))
    im2(:, :) = im(:, :) / 2;
    t = toc();
    disp(t);
    im = nearest_uint8(im2)
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function noise_01_button()
    global message_text;
    p = 0.02;
    message_text.String = string(p) + "% の確率でピクセルが変化";
    disp("noise_01_button");
    im = imnoise(im, 'salt & pepper', p);
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction
    
function noise_gauss_button()
    global message_text;
    p = 0.10;
    message_text.String = "Gaussian noise";
    disp("noise_gauss_button");
    im = imnoise(im, 'gaussian');
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function block_shuffle_button()
    global message_text;
    message_text.String = "<html>画像をタイル分割してシャッフルしました．<br>画像行列 <i>X</i>を置換行列 <i>P</i> を用いて <i>PXP</i><sup>T</sup> と変換しました．</html>";
    disp("block_shuffle_button");
    n_block = int(shuffle_slider.Value);
    sizexy = size(im);
    block_size =  int(sizexy / n_block);
    x_index = grand(1, "prm", 1 : n_block);
    y_index = grand(1, "prm", 1 : n_block);
    disp(n_block, block_size, x_index, y_index);
    im2 = im;
    for i = 1 : n_block do
        for j = 1 : n_block do
            im2((i - 1) * block_size(1) + (1 : block_size(1)), (j - 1) * block_size(2) + (1 : block_size(2))) = im((x_index(i) - 1) * block_size(1) + (1 : block_size(1)), (y_index(j) - 1) * block_size(2) + (1 : block_size(2)))
        end
    end
    im = im2;
    imshow(im);
    [im, message_text] = resume(im, message_text); // Write back
endfunction

function update_shuffle_slider(slider)
    global shuffle_slider_text;
    disp("update shuffle slider");
    shuffle_slider.Value = int(shuffle_slider.Value);
    disp(shuffle_slider.Value);
    shuffle_slider_text.String = string(shuffle_slider.Value) + shuffle_slider_text_suffix;
    [im, message_text] = resume(im, message_text); // Write back
endfunction

// Parameters
WINDOW_SIZE_X = 800;
WINDOW_SIZE_Y = 800;
MESSAGE_POS_Y = 0.6;
UI_FONT_SIZE = 20;

// Run commands
stacksize(30000000);
global im;
global targe_image;
im = [1, 0; 0, 1];

number_of_button = 8;
BUTTON_SIZE_X = 1.0 / 3.0;
BUTTON_SIZE_Y = MESSAGE_POS_Y / number_of_button;

h = createWindow();
h.Position = [0, 0, WINDOW_SIZE_X, WINDOW_SIZE_Y];

// Bottom to top
for i = 1 : number_of_button do
    huibutton(i) = uicontrol(h, "style", "pushbutton"); 
    huibutton(i).Units = "normalized";
    huibutton(i).Position = [0, MESSAGE_POS_Y - BUTTON_SIZE_Y * i, BUTTON_SIZE_X, BUTTON_SIZE_Y];
    huibutton(i).fontsize = UI_FONT_SIZE;
end

ii = 1;
huibutton(ii).String = "終了";
huibutton(ii).Callback = "quit_button";

ii = ii + 1;
huibutton(ii).String = "画像の選択";
huibutton(ii).Callback = "select_image_button";

ii = ii + 1;
huibutton(ii).String = "画像を最初に戻す";
huibutton(ii).Callback = "set_default_button";

ii = ii + 1;
huibutton(ii).String = "画像を暗くする";
huibutton(ii).Callback = "dark_button";

ii = ii + 1;
huibutton(ii).String = "画像を白黒反転";
huibutton(ii).Callback = "imcomplement_button";

ii = ii + 1;
huibutton(ii).String = "2次のガウスぼかし";
huibutton(ii).Callback = "gaussian_2_button";

ii = ii + 1;
SHUFFLE_ROW_NUM = ii;
huibutton(ii).String = "画像のタイルシャッフル";
huibutton(ii).Callback = "block_shuffle_button";

shuffle_slider = uicontrol(h, "style", "slider"); 
shuffle_slider.Units = "normalized";
shuffle_slider.Position = [BUTTON_SIZE_X, MESSAGE_POS_Y - BUTTON_SIZE_Y * SHUFFLE_ROW_NUM, BUTTON_SIZE_X, BUTTON_SIZE_Y];
shuffle_slider.Min = 4;
shuffle_slider.Max = 12;
shuffle_slider.Value = 8; // default value
shuffle_slider.SliderStep = [1, 2]; // [small, big]
shuffle_slider.Callback = "update_shuffle_slider"; // Called when slider value updated

shuffle_slider_text = uicontrol(h, "style", "text"); 
shuffle_slider_text.Units = "normalized";
shuffle_slider_text_suffix = "分割タイルを使用する";
shuffle_slider_text.Position = [BUTTON_SIZE_X * 2, MESSAGE_POS_Y - BUTTON_SIZE_Y * SHUFFLE_ROW_NUM, BUTTON_SIZE_X, BUTTON_SIZE_Y];
shuffle_slider_text.fontsize = UI_FONT_SIZE;
shuffle_slider_text.String = string(shuffle_slider.Value) + shuffle_slider_text_suffix; // Japanese text cannot be used.
shuffle_slider_text.HorizontalAlignment = "center"

ii = ii + 1;
SVD_ROW_NUM = ii;
huibutton(SVD_ROW_NUM).String = "画像を圧縮（特異値分解）";
huibutton(SVD_ROW_NUM).Callback = "svd_slider_percent_button";

svd_slider = uicontrol(h, "style", "slider"); 
svd_slider.Units = "normalized";
svd_slider.Position = [BUTTON_SIZE_X, MESSAGE_POS_Y - BUTTON_SIZE_Y * SVD_ROW_NUM, BUTTON_SIZE_X, BUTTON_SIZE_Y];
svd_slider.Min = 1;
svd_slider.Max = 25; // R
svd_slider.Value = 10; // default value
svd_slider.SliderStep = [1, 10]; // [small, big]
svd_slider.Callback = "update_svd_slider"; // Called when slider value updated

svd_slider_text = uicontrol(h, "style", "text"); 
svd_slider_text.Units = "normalized";
svd_slider_text_suffix = "% にデータサイズを圧縮";
svd_slider_text.Position = [BUTTON_SIZE_X * 2, MESSAGE_POS_Y - BUTTON_SIZE_Y * SVD_ROW_NUM, BUTTON_SIZE_X, BUTTON_SIZE_Y];
svd_slider_text.fontsize = UI_FONT_SIZE;
svd_slider_text.String = string(svd_slider.Value * 2) + svd_slider_text_suffix; // Japanese text cannot be used.
svd_slider_text.HorizontalAlignment = "center"


message_text = uicontrol(h, "style", "text"); 
message_text.Units = "normalized";
message_text.Position = [0, MESSAGE_POS_Y, 1.0, 1.0 - MESSAGE_POS_Y];
message_text.fontsize = UI_FONT_SIZE;
message_text.HorizontalAlignment = "center"
message_text.String = "<html>ここに簡単な解説が表示されます．</html>";
message_text.BackgroundColor = [1.0, 1.0, 1.0];

